//	FileName: core.v
//	Unified Sram Core Model for Sync Burst/NBT/DDR Sram 
//	Revision History:
//	7/12/00	 1.0  1) first draft, DDR is not implemented yet
//	7/13/00	 1.1  1) reduce state machine to 3 states
//	7/14/00	 1.2  1) add task-shiftnow for suspend cycle
//	7/16/00	 1.3  1) add NBT function
//		      2) reduce ZZ pesmistics at powerup
//		      3) add nop(NBT) cycle - dummy W (BW=0)is nop 
//	7/17/00	 1.4  1) add the NBT R/W same address match case
//	5/14/01	 1.5  2) removed jtag info, added W and ADV back in
//      5/31/01  1.6  1) corrected lw and llw mode for the NBT.
//                       a) lw is for NBT Flow
//                       b) llw is for NBT Pipe
//                    2) Changed ce assgin block for SCD addition
//                    3) Changed oe assgin block for SCD addition
//                    4)  Added ce2 signal for SCD output hold
//      8/27/01  1.7  1) Fixed a race condition between = and <= assginments
//	              2) In the state Machine section, shiftnow and all the states
//      9/25/01  1.8  1) Shorten the DCD data hold on a deselect.  needed to 
//	              2) shorten by one complete cycle
//      4/04/02  1.9  1) Model did not support a Suspend Burst command changed
//                    2) ce default from (ce0 | ce1) to (ce0)
//                    3) Suspend command in state table not outputs last read command data
//      6/25/02  2.0  1) Changed EXTEST command from 001 to 100
//     02/20/03  2.1  1) Modified read to look at previous write byte write's and 
//			 determine where to get the data from.
//     08/12/03  2.2  1) Added a DELAY variable that allows the use of tKQX to simulate output hold
//                    2) updated state maching to seperate the Burst Read and Burst Write states.
//                    3) updated internal Bytewrite signals for NBT mode, they ignore the nW
//     03/16/04  2.3  1) updated state machine to have a suspend state.
//                    2) Changed idle state to deselect.
//     04/28/04   2.4 1) Rearranged state that determins Deselect, Burst and Suspend
//---------------------------------------------------------------
//  	Preload SRAM For Debugging
//---------------------------------------------------------------
    reg     [DQ_size-1:0]   bank0 [0:bank_size],
                            bank1 [0:bank_size],
                            bank2 [0:bank_size],
                            bank3 [0:bank_size],
                            bank4 [0:bank_size],
                            bank5 [0:bank_size],
                            bank6 [0:bank_size],
                            bank7 [0:bank_size];
	//---------------------------------------------------------------
	//      Valid Sync Sram Operatons  - 3 configuration pins 
	//---------------------------------------------------------------

    wire    [2:0]           mode = {nFT,NBT,DDR}; 
				// 000 - flowthrough burst
				// 100 - pipeline burst
				// 010 - flowthrough NBT
				// 110 - pipeline NBT
				// 001 - flowthrough DDR burst
				// 101 - pipeline DDR burst
				// 011 - flowthrough DDR NBT
				// 111 - pipeline DDR NBT

	//---------------------------------------------------------------
    //		NBT Sram & Burst Sram External Pins 
	//---------------------------------------------------------------
//  Burst Sram Unique Input Pins 
	wire				nADSC;	// processor address strobe  
	wire				nADSP; 	// cache address strobe
	wire				nADV;	// ADV (negative active)
	wire				nGW;	// global write enable
	wire				nBW;	// byte write selection
//  NBT Sram Unique Input Pins
	wire				nCKE;	// clock enable
	wire				nW;	// write strobe
	wire				pADV;	// ADV (positive active) 

//	Setting below is based on the actual chip,
//	When these pins are left floating, the chip is internally set to: 
//	=> No sleep, Pipeline , Interleaved Burst Address 
//    => Single cycle deselect, Even parity, Parity enable
    
	pulldown(ZZ);
	pullup(nFT);
	pullup(nLBO);
	pullup(SCD);
	pullup(DP);
	pulldown(nPE);
	pullup(E2);
	pulldown(nE3);

	//---------------------------------------------------------------
	//  	Merging Input Pins of NBT / Burst Operation
	//---------------------------------------------------------------

	wire					ew,lw,llw,ft,pl,pld; 	// internal read write 
	wire					CKE;	// internal clock enable	
	wire					E;	// internal chip enable
	wire					W;	// internal write strobe
	wire	[7:0]				BW;	// internal byte write enable
	wire					ADV;	// internal address advance
	wire					ADS;	// internal address strobe
	wire					ADSP;	// internal ADSP 
	wire					ZNOP;	// NBT mode the nop cycle
                  real                                                                                  DELAY;

   
	assign	ew 	= ~mode[1];			// early write
	assign	lw 	=  mode[1] & ~mode[0] & ~nFT;	// late write
	assign	llw	=  mode[1] & ~mode[0] &  nFT;	// late late write
	assign	ft 	=  ~mode[2];			// flowthrough read
	assign	pl 	=  mode[2];			// pipeline read

	assign	E  	= ~nE1  & E2 & ~nE3;
	assign	CKE 	=  mode[1] ? ~nCKE : 1;
	assign	ADSP	=  mode[1] ? 0 : (~nADSP & ~nE1); 
	assign	ADV	=  mode[1] ?  pADV : ( ~nADV & ~ADSP & nADSC); 
	assign	ADS	=  mode[1] ? ~pADV : ( ADSP | ~nADSC); 
	assign	W	=  mode[1] ? (~nW & (~nBa|~nBb|~nBc|~nBd|~nBe|~nBf|~nBg|~nBh)) : 
			   (~ADSP & (~nGW | (~nBW & (~nBa|~nBb|~nBc|~nBd|~nBe|~nBf|~nBg|~nBh))));  
	assign 	BW[0]	=  mode[1] ? ( ~nBa) : (~nGW | (~nBW & ~nBa));
	assign 	BW[1]	=  mode[1] ? ( ~nBb) : (~nGW | (~nBW & ~nBb));
	assign 	BW[2]	=  mode[1] ? ( ~nBc) : (~nGW | (~nBW & ~nBc));
	assign 	BW[3]	=  mode[1] ? ( ~nBd) : (~nGW | (~nBW & ~nBd));
	assign 	BW[4]	=  mode[1] ? ( ~nBe) : (~nGW | (~nBW & ~nBe));
	assign 	BW[5]	=  mode[1] ? ( ~nBf) : (~nGW | (~nBW & ~nBf));
	assign 	BW[6]	=  mode[1] ? ( ~nBg) : (~nGW | (~nBW & ~nBg));
	assign 	BW[7]	=  mode[1] ? ( ~nBh) : (~nGW | (~nBW & ~nBh));
	assign  ZNOP	=  mode[1] & ~nW & (BW==0);

	//---------------------------------------------------------------
	//  	Gated SRAM Clock
	//---------------------------------------------------------------
      
    reg                     clk_i;
    always begin
        @ ( posedge CK ) begin
                    clk_i = CKE & ~ZZ;
            end
        @ ( negedge CK ) begin
                    clk_i = 1'b0;
            end
    end


	//---------------------------------------------------------------
	//
	//  	Combinatorial Logic 
	//
	//---------------------------------------------------------------

	// Internal data is always 9 bits
	reg     [DQ_size-1:0]   Qai,Qbi,Qci,Qdi,Qei,Qfi,Qgi,Qhi; // read data
	reg     [DQ_size-1:0]   Dai,Dbi,Dci,Ddi,Dei,Dfi,Dgi,Dhi; // write data 

	reg	[7:0]		bw0; 	// saved bw 
	reg	[7:0]		bw1; 	// bw buffer 1 
	reg	[7:0]		bw2; 	// bw buffer 2 
	reg	[A_size-1:0] 	addr0; 	// saved address 
	reg	[A_size-1:0] 	addr1;	// address buffer 1
	reg	[A_size-1:0] 	addr2;	// address buffer 2

	reg 	[1:0]           bcnt; 	// burst counter 
	wire	[A_size-1:0] 	baddr; 	// burst memory address
	wire	[A_size-1:0] 	waddr; 	// write memory address
	wire	[A_size-1:0] 	raddr; 	// read memory address

	// output of state machine
	reg                 	we0, we1, we2;	// we flags 
	reg                 	re0, re1, re2;	// re flags
	reg			ce0, ce1, ce2;	// ce flags

	// 
	wire                 	ce, oe, we, re;	// active ones 
	wire	[7:0]		bw ; 		// active ones

	assign baddr	= nLBO ? {addr0[A_size-1:2],bcnt[1]^addr0[1],bcnt[0]^addr0[0]} : 
	               {addr0[A_size-1:2], addr0[1:0]+bcnt} ; 
	assign waddr	= ew ? baddr : ( lw ? addr1 : addr2);
	assign raddr	= ft ? baddr : ( pl ? addr1 : addr2);
	assign bw	= ew ? bw0 : ( lw  ? bw1 : bw2);
	assign we	= ew ? we0 : ( lw  ? we1 : we2);
	assign re	= ft ? re0 : ( pl ? re1 : re2);
//	assign ce 	= (~SCD & nFT) & re2==1 ? ce1 : ~SCD ? ce1 : (ce0 | ce1);
//	assign oe 	= (~SCD & nFT) & re2==1 ? ce1 : ~SCD & re1==1 ? ce1 : ((mode[1] ? 1: ~we) & re & ce);
	assign ce 	= (~SCD) & re2==1 ? ce1 : (mode[1] ? ce1: ce0);
//	assign ce 	= (~SCD ) & re2==1 ? ce1 : ~SCD ? ce1 : (ce0);
	assign oe 	= (~SCD) & re2==1 ? ce1 : ((mode[1] ? 1: ~we) & re & ce);

   

    //---------------------------------------------------------------
    //
    //              Synchronuos Logic 
    //
    //---------------------------------------------------------------
	parameter        	deselect	= 0,
                     	newcyc	= 1,
                    	burst	= 2,
		suspend = 3;

	reg 	[2:0]  		state;

	initial
	  begin
	     state <= deselect;
	     ce0 <= 0; ce1 <= 0; ce2 <= 0;
	     we0 <= 0; we1 <= 0; we2 <= 0;
	     re0 <= 0; re1 <= 0; re2 <= 0;
	     bw0 <= 0; bw1 <= 0; bw2 <= 0; 
	     bcnt=0;
	  end
		

	//---------------------------------------------------------------
	//  	State Machine
	//---------------------------------------------------------------

   task shiftnow;
      begin
	 addr2 <= addr1;
	 addr1 <= baddr;
	 bw2 <= bw1; bw1 <= bw0; 
	 we2 <= we1; we1 <= we0;
	 re2 <= re1; re1 <= re0;
         ce2 <= ce1; ce1 <= ce0; 
	 //
	 //		save data bus for later write 
	 //
	 Dai <= DQa;
	 Dbi <= DQb;
	 Dci <= DQc;
	 Ddi <= DQd;
	 Dei <= DQe;
	 Dfi <= DQf;
	 Dgi <= DQg;
	 Dhi <= DQh;
      end
   endtask

   always @ (posedge clk_i) begin: st
      case (state)
        deselect: 
	  begin
	     if(E==1 & ADS==1) begin
		shiftnow;
		state <= newcyc;
		we0<=W; re0<=~W & ~ZNOP; ce0<=1;
		addr0<=A; bw0<=BW;
		bcnt<=0;
	     end
	     if( E!=1 & ADS==1) begin
		shiftnow;
		state <= deselect;
		we0<=0; re0<=0; ce0<=0;
	     end
	  end
	
        newcyc:
	  begin 
	     if(E==1 & ADS==1) begin
		shiftnow;
		state <= newcyc;
		we0<=W; re0<=~W & ~ZNOP; ce0<=1;
		addr0<=A; bw0<=BW;
		bcnt<=0;
	     end
	     if(( E!=1 | ADS==0)) begin 
		if( E!=1 & ADS==1) begin
		   shiftnow;
		   state <= deselect;
		   we0<=0; re0<=0; ce0<=0;
		end
		else if(( E!=1 | ADS==0) & ADV==0) begin 
		   // suspend
		   shiftnow;
		   we0<=W; re0<=~W & ~ZNOP; ce0<=1;
		   addr0<=A; bw0<=BW;
		   if (W==1) state <= burst;
		   else      state <= suspend;
		end
	     end // if (( E!=1 | ADS==0))
	     if(ADS==0 & ADV==1) begin
		// continue
		if (we0==1) begin
		   shiftnow;
		   state <= burst;
		   we0<=ADV; ce0<=1;
		   bw0<=BW;
		   bcnt<= bcnt+1;
		end
		else if (re0==1) begin
		   shiftnow;
		   state <= burst;
		   re0<=ADV & ~ZNOP; ce0<=1;
		   bcnt<= bcnt+1;
		end
	     end // if (ADS==0 & ADV==1)
	  end // case: newcyc
	
        burst: 
	  begin 
	     if(E==1 & ADS==1) begin
		shiftnow;
		state <= newcyc;
		we0<=W; re0<=~W & ~ZNOP; ce0<=1;
		addr0<=A; bw0<=BW;
		bcnt<=0;
	     end
	     if(( E!=1 | ADS==0)) begin 
		if( E!=1 & ADS==1) begin
		   shiftnow;
		   state <= deselect;
		   we0<=0; re0<=0; ce0<=0;
		end
		else if(( E!=1 | ADS==0) & ADV==0) begin 
		   // suspend
		   shiftnow;
		   we0<=W; re0<=~W & ~ZNOP; ce0<=1;
		   addr0<=A; bw0<=BW;
		   if (W==1) state <= burst;
		   else      state <= suspend;
		end
	     end // if (( E!=1 | ADS==0))
	     if(ADS==0 & ADV==1) begin
		if (we0==1) begin
		   // continue 
		   shiftnow;
		   state <= burst;
		   we0<=ADV; ce0<=1;
		   bw0<=BW;
		   bcnt<= bcnt+1;
		end
		  else if(re0==1) begin
		     // continue 
		     shiftnow;
		     state <= burst;
		     re0<=ADV & ~ZNOP; ce0<=1;
		     bcnt<= bcnt+1;
		  end
	     end // if (ADS==0 & ADV==1)
	  end // case: burst
	
	suspend:
	  begin
	     if(E==1 & ADS==1) begin
		shiftnow;
		state <= newcyc;
		we0<=W; re0<=~W & ~ZNOP; ce0<=1;
		addr0<=A; bw0<=BW;
		bcnt<=0;
	     end
	      else if( E!=1 | ADS==0) begin
		if(E!=1 & ADS==1) begin
		   shiftnow;
		   state <= deselect;
		   we0<=0; re0<=0; ce0<=0;
		end
		else if (( E!=1 | ADS==0) & ADV==0) begin
 		   // suspend
		   shiftnow;
		   we0<=W; re0<=~W & ~ZNOP; ce0<=1;
		   addr0<=A; bw0<=BW;
		   if (W==1) state <= newcyc;
		   else      state <= suspend;
		end
	     end // if (~E==1 | ADS==0)
	     else if(ADS==0 & ADV==1) begin
		if (we0==1) begin
		   // continue 
		   shiftnow;
		   state <= burst;
		   we0<=ADV; ce0<=1;
		   bw0<=BW;
		   bcnt<= bcnt+1;
		end
		else if(re0==1) begin
		   // continue 
		   shiftnow;
		   state <= burst;
		   re0<=ADV & ~ZNOP; ce0<=1;
		   bcnt<= bcnt+1;
		end
	     end // if (ADS==0 & ADV==1)
	  end // case: suspend
	
	default:
	  begin
	     shiftnow;
	     state <= deselect;
	     we0<=0; re0<=0; ce0<=0;
	     bcnt<=0;
	  end
      endcase

    //---------------------------------------------------------------
    //
    //              Data IO Logic 
    //
    //---------------------------------------------------------------


   #0 	if (we) begin 
      if (bw[0]==1)	bank0[waddr] <= Dai;
      if (bw[1]==1)	bank1[waddr] <= Dbi;
      if (bw[2]==1)	bank2[waddr] <= Dci;
      if (bw[3]==1)	bank3[waddr] <= Ddi;
      if (bw[4]==1)	bank4[waddr] <= Dei;
      if (bw[5]==1)	bank5[waddr] <= Dfi;
      if (bw[6]==1)	bank6[waddr] <= Dgi;
      if (bw[7]==1)	bank7[waddr] <= Dhi;
      
   end
   
   //
   // 		reading from the ram cell
   //
   #1									// delay is necessary for W to complete
     // before R can take place 
     if(we==0 | raddr != waddr) begin	// 7-17 R & W address different 
	Qai <= bank0[raddr];
	Qbi <= bank1[raddr];
	Qci <= bank2[raddr];
	Qdi <= bank3[raddr];
	Qei <= bank4[raddr];
	Qfi <= bank5[raddr];
	Qgi <= bank6[raddr];
	Qhi <= bank7[raddr];
     end
     else begin							// 7-17 NBT W & R address match 
	if (bw[0]==1) Qai <= Dai;
	else          Qai <= bank0[raddr];
	if (bw[1]==1) Qbi <= Dbi;
	else          Qbi <= bank1[raddr];
	if (bw[2]==1) Qci <= Dci;
	else          Qci <= bank2[raddr];
	if (bw[3]==1) Qdi <= Ddi;
	else          Qdi <= bank3[raddr];
	if (bw[4]==1) Qei <= Dei;
	else          Qei <= bank4[raddr];
	if (bw[5]==1) Qfi <= Dfi;
	else          Qfi <= bank5[raddr];
	if (bw[6]==1) Qgi <= Dgi;
	else          Qgi <= bank6[raddr];
	if (bw[7]==1) Qhi <= Dhi;
	else          Qhi <= bank7[raddr];
     end
   end 	// st

	//---------------------------------------------------------------
	//
	// 	Data Output
	//                          
	//---------------------------------------------------------------

   always @ (re0)
     begin
	DELAY        <= ~re0 ? tKQX: (tKQ-1);
     end
   
	parameter	HighZ = 9'b ZZZZZZZZZ;

	assign		#(DELAY) DQa = (nG|~oe|ZZ) ? HighZ : Qai;  // 
	assign		#(DELAY) DQb = (nG|~oe|ZZ) ? HighZ : Qbi;  // 
	assign		#(DELAY) DQc = (nG|~oe|ZZ) ? HighZ : Qci;  // 
	assign		#(DELAY) DQd = (nG|~oe|ZZ) ? HighZ : Qdi;  // 
	assign		#(DELAY) DQe = (nG|~oe|ZZ) ? HighZ : Qei;  // 
	assign		#(DELAY) DQf = (nG|~oe|ZZ) ? HighZ : Qfi;  // 
	assign		#(DELAY) DQg = (nG|~oe|ZZ) ? HighZ : Qgi;  // 
	assign		#(DELAY) DQh = (nG|~oe|ZZ) ? HighZ : Qhi;  // 